// Copyright 2017 The Johns Hopkins University Applied Physics Laboratory.
// Licensed under the MIT License. See LICENSE.txt in the project root for full license information.


#pragma once

#include <cmath>
#include <vector>
#include <iostream>
#include <utility>
#include <iterator>
#include <limits>
#include <fstream>

#include <eopcc/utility.hpp>

namespace eopcc {

/// RGB or HSV color of a point
template <typename T>
struct PointColor {
    union {
        T red;
        T hue;
    };
    union {
        T green;
        T sat;
    };
    union {
        T blue;
        T val;
    };

    T intensity() {
        return (red + green + blue)/3.0;
    }
};




/// Structure holding points and their intensities, colors, or metadata
struct PointCloud {
	std::string wkt_spatial_ref;
  Extent extent;

  std::vector<Point> loc;
  std::vector<intensity_t> intensity;

  bool has_color;
  std::vector<PointColor<color_t>> color;

  bool has_time;
  std::vector<double> gps_time;

  bool has_las_metadata;
  std::vector<LASMetadata> las_metadata;

  PointCloud() {
    has_time = true;
    has_color = true;
    has_las_metadata = false;
  }

	/// Resizes the point vector and other active vectors
  void resize(size_t num_points) {
    loc.resize(num_points);
    intensity.resize(num_points);

    if (has_color) {
      color.resize(num_points);
    }

    if (has_time) {
      gps_time.resize(num_points);
    }

    if (has_las_metadata) {
      las_metadata.resize(num_points);
    }
  }

	/// Returns the current size
  size_t size() const {
    return loc.size();
  }


	/// Appends all the points in other PointCloud to this PointCloud
	void append(PointCloud& other);

	/// Sets a point in this cloud from the values in other at other_index
	void set_from(size_t index, const PointCloud& other, int other_index);

	/// Creates a subcloud from a range of indices
	PointCloud subcloud(size_t sub_start, size_t sub_end) const;

	// Calculates and returns the mean X/Y/Z location of this cloud
	Point mean() const;

	// Translates this cloud by the X/Y/Z values of a given point. Updates the extent
	void translate(const Point& translation);

  // Given a vector of keep/trash flags creates a new vector with only the keep members
  PointCloud filter(const std::vector<bool>& filter_vector) const;
  void filter_inplace(const std::vector<bool>& filter_vector);

	// Calculates the 3D bounding box (extent) of this cloud (but does not modify the extent member variable)
	Extent calculate_extent() const;

	// Updates the extent member with the new extent
  void update_extent() {
    extent = calculate_extent();
  }

	// Counts the points inside the given extent
  int inside_extent(const Extent& extent);
  int inside_extent(const PolygonExtent& extent);
};

/// Generate and return a point cloud with random positions uniformly distributed over the extent
PointCloud RandomCloud(size_t number_points, const Extent& extent);
PointCloud GenerateSample(std::string kml_file, double spacing);



} // namespace eopcc
